/////////////////////////////////////////////////////////////////////////////
// Name:        wx/printcustom.h
// Purpose:     Support classes for creating custom print dialogs
// Author:      Hans de Ruiter
// Modified by:
// Created:     03/11/2010
// RCS-ID:      $ID: printcustom.h $
// Copyright:   (c) Hans de Ruiter
// Licence:     wxWindows Licence
/////////////////////////////////////////////////////////////////////////////

#ifndef _WX_PRINTCUSTOM_H_BASE_
#define _WX_PRINTCUSTOM_H_BASE_

#include "wx/defs.h"

#if wxUSE_PRINTING_ARCHITECTURE

#include "wx/event.h"
#include "wx/dialog.h"
#include "wx/intl.h"
#include "wx/cmndata.h"
#include "wx/paper.h"
#include "wx/sharedptr.h"
#include "wx/string.h"
#include "wx/vector.h"

// ---------------------------------------------------------------------------
// wxNativePrinterInfo: Native printer information for wxPrinterInfo
// ---------------------------------------------------------------------------

class wxNativePrinterInfo
{
public:
	/** Creates a new wxNativePrinterInfo object.
	 *
	 * @param isDefault set to true if this is the default printer, and
	 * false otherwise
	 */
	wxNativePrinterInfo(bool isDefault);

	virtual ~wxNativePrinterInfo();

	/** Returns a pointer to the database of paper sizes that this printer
	 * supports
	 *
	 * @return wxSharedPtr<wxPrintPaperDatabase> pointer to the printer database
	 */
	virtual wxSharedPtr<wxPrintPaperDatabase> GetSupportedPaperSizes() const = 0;

	/** Gets the non-printable margins in tenths of a millimetre
	 * // ##### FIXME! ##### This assumes that the non-printable margin is the same regardless of paper size. Is this a valid assumption?
	 * 
	 * @param leftMargin the size of the left margin is written to this parameter
	 * @param rightMargin the size of the right margin is written to this parameter
	 * @param topMargin the size of the top margin is written to this parameter
	 * @param bottomMargin the size of the bottom margin is written to this parameter
	 */
	virtual void getNonPrintableMargins(
		unsigned int *leftMargin, unsigned int *rightMargin,
		unsigned int *topMargin, unsigned int *bottomMargin) const = 0;

	/** Returns true if the given printer data object uses this printer.
	 */
	virtual bool IsSamePrinter(const wxPrintData &printData) const = 0;

	/** Returns a wxPrintData object for this printer, and the default page
	 * settings.
	 */
    virtual const wxPrintData GetPrintData() = 0;

	/** Updates the given print data object to use this printer. 
	 * This will try to choose the closest comparable paper type that this printer
	 * supports.
	 *
	 * @param printData the print data object to update
	 */
	virtual void UpdatePrintData(wxPrintData &printData) const = 0;

	/** Sets a wxPrintData's paper type to the one specified.
	 * IMPORTANT: Use this instead of calling wxPrintData::SetPaperId(). This
	 * method will correctly handle printer-specific paper IDs that are only in
	 * this printer's database. 
	 *
	 * @param paperID the paper ID. This ID should be in this printer's database
	 * @param printData the printData object whose paper type will be set
	 *
	 * @return bool true if successful; false if the Paper ID wasn't found
	 */
	virtual bool SetPrintDataPaperType(wxPaperSize paperID, wxPrintData &printData) const = 0;

	/** Returns true if this is the default printer.
	 */
	virtual bool IsDefaultPrinter() const;

protected:
	/** The database of supported paper sizes.
	 * IMPORTANT: Child-classes should create this in its constructor.
	 */
	wxSharedPtr<wxPrintPaperDatabase> paperDatabase;

	bool m_isDefault;
};

// ---------------------------------------------------------------------------
// wxPrinterInfo: Enables getting basic information about a printer's 
// capabilities
// ---------------------------------------------------------------------------

class WXDLLIMPEXP_CORE wxPrinterInfo : public wxObject
{
public:
	/** Constructs a wxPrinterInfo object for a given wxNativePrinterInfo 
	 * object.
	 *
	 * @param nativePrinterInfo pointer to the native printer info object
	 */
    wxPrinterInfo(wxSharedPtr<wxNativePrinterInfo> nativePrinterInfo);
    
	virtual ~wxPrinterInfo();

	/** Returns a pointer to the database of paper sizes that this printer
	 * supports
	 *
	 * @return wxSharedPtr<wxPrintPaperDatabase> pointer to the printer database
	 */
	wxSharedPtr<wxPrintPaperDatabase> GetSupportedPaperSizes();

	/** Gets the non-printable margins in tenths of a millimetre
	 * // ##### FIXME! ##### This assumes that the non-printable margin is the same regardless of paper size. Is this a valid assumption?
	 * 
	 * @param leftMargin the size of the left margin is written to this parameter
	 * @param rightMargin the size of the right margin is written to this parameter
	 * @param topMargin the size of the top margin is written to this parameter
	 * @param bottomMargin the size of the bottom margin is written to this parameter
	 */
	void getNonPrintableMargins(unsigned int *leftMargin, unsigned int *rightMargin,
		unsigned int *topMargin, unsigned int *bottomMargin);

	/** Returns true if the given printer data object uses this printer.
	 */
	bool IsSamePrinter(const wxPrintData &printData) const;

	/** Returns a wxPrintData object for this printer, and the default page
	 * settings.
	 */
    wxPrintData GetPrintData();

	/** Updates the given print data object to use this printer. 
	 * This will try to choose the closest comparable paper type that this printer
	 * supports.
	 *
	 * @param printData the print data object to update
	 */
	void UpdatePrintData(wxPrintData &printData) const;

	/** Sets a wxPrintData's paper type to the one specified.
	 * IMPORTANT: Use this instead of calling wxPrintData::SetPaperId(). This
	 * method will correctly handle printer-specific paper IDs that are only in
	 * this printer's database. 
	 *
	 * @param paperID the paper ID. This ID should be in this printer's database
	 * @param printData the printData object whose paper type will be set
	 *
	 * @return bool true if successful; false if the Paper ID wasn't found
	 */
	bool SetPrintDataPaperType(wxPaperSize paperID, wxPrintData &printData) const;

	/** Returns true if this is the default printer.
	 */
	bool IsDefaultPrinter() const;

private:
    wxSharedPtr<wxNativePrinterInfo> m_nativePrinterInfo;

	wxSharedPtr<wxPrintPaperDatabase> m_paperDatabase;
    
private:
    wxDECLARE_NO_COPY_CLASS(wxPrinterInfo);
};


// ---------------------------------------------------------------------------
// wxPrinterSysInfo: Gets a list of available printers
// ---------------------------------------------------------------------------

class WXDLLIMPEXP_CORE wxPrinterSysInfo
{
public:
	/** Returns a vector of all available printers.
	 */
    static wxVector<wxSharedPtr<wxPrinterInfo>> GetAvailablePrinters();

	/** Returns a pointer to the default printer.
	 */
	static wxSharedPtr<wxPrinterInfo> GetDefaultPrinter();

	/** Returns a pointer to the printer selected by the given print data object.
	 *
	 * @param printData the printData object containing the printer settings
	 *
	 * @return wxSharedPtr<wxPrinterInfo> pointer to the wxPrinterInfo that matches
	 * the printData (or NULL if nothing was found)
	 */
	static wxSharedPtr<wxPrinterInfo> GetPrinterByPrintData(const wxPrintData* printData);

private:
	/** The global vector of available printers.
	 */
	static wxVector<wxSharedPtr<wxPrinterInfo>> availablePrinters;
};

// ---------------------------------------------------------------------------
// wxNativePrinterPrefsDialog: Base class for native implementations of 
// wxPrinterPrefsDialog
// ---------------------------------------------------------------------------

class wxNativePrinterPrefsDialog
{
public:
	wxNativePrinterPrefsDialog(){};

	virtual ~wxNativePrinterPrefsDialog(){};

    virtual int ShowModal() = 0;

    virtual wxPrintData& GetPrintData() = 0;
};

// ---------------------------------------------------------------------------
// wxPrinterPrefsDialog: Shows the OS printer preferences dialog
// ---------------------------------------------------------------------------

class WXDLLIMPEXP_CORE wxPrinterPrefsDialog : public wxObject
{
public:
	/** Creates a printer prefs dialog.
	 *
	 * @param parent pointer to the parent window
	 * @param data pointer to the current print data. If NULL, then 
	 * the default printer and settings will be used.
	 */
    wxPrinterPrefsDialog(wxWindow *parent, wxPrintData *data = NULL);

    virtual ~wxPrinterPrefsDialog();

    int ShowModal();

    wxPrintData& GetPrintData();

private:
    wxScopedPtr<wxNativePrinterPrefsDialog> m_nativeDialog;
    
private:
    wxDECLARE_NO_COPY_CLASS(wxPrinterPrefsDialog);
};


#endif

#endif
    // _WX_PRINTCUSTOM_H_BASE_