/////////////////////////////////////////////////////////////////////////////
// Name:        wx/msw/prntcustom.cpp
// Purpose:     MS Windows implementation of support classes for creating 
//              custom print dialogs
// Author:      Hans de Ruiter
// Modified by:
// Created:     03/11/2010
// RCS-ID:      $ID: msw/prntcustom.cpp $
// Copyright:   (c) Hans de Ruiter
// Licence:     wxWindows Licence
/////////////////////////////////////////////////////////////////////////////


// ===========================================================================
// declarations
// ===========================================================================

// ---------------------------------------------------------------------------
// headers
// ---------------------------------------------------------------------------

// For compilers that support precompilation, includes "wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#include "wx/msw/prntcustom.h"
#include "wx/msw/printdlg.h"

// Don't use the Windows printer if we're in wxUniv mode and using
// the PostScript architecture
#if wxUSE_PRINTING_ARCHITECTURE && (!defined(__WXUNIVERSAL__) || !wxUSE_POSTSCRIPT_ARCHITECTURE_IN_MSW)

#ifndef WX_PRECOMP
    #include "wx/msw/wrapcdlg.h"
    #include "wx/window.h"
    #include "wx/msw/private.h"
    #include "wx/utils.h"
    #include "wx/dc.h"
    #include "wx/app.h"
    #include "wx/msgdlg.h"
    #include "wx/intl.h" 
    #include "wx/log.h"
    #include "wx/dcprint.h"
    #include "wx/dcmemory.h"
    #include "wx/image.h"
#endif


#include <stdlib.h>

#ifndef __WIN32__
    #include <print.h>
	#include <windows.h>
	#include <tchar.h>
#endif

#define TENTHSMM_TO_INCHES(x) ((x) * 0.00393700787)
#define INCHES_TO_TENTHSMM(x) ((x) / 0.00393700787)
#define PAPERNAME_LENGTH 64
#define wxPAPER_USER 200 // ##### FIXME! ##### wxWidgets really should provide a method of adding custom printer-specific paper types to the database



// ===========================================================================
// implementation
// ===========================================================================

// ---------------------------------------------------------------------------
// Printer System
// ---------------------------------------------------------------------------

wxVector<wxSharedPtr<wxPrinterInfo>> wxPrinterSysInfo::GetAvailablePrinters()
{
	if(availablePrinters.size() > 0)
	{
		// Already have the list
		return availablePrinters;
	}

	DWORD dwNeeded   = 0;
	DWORD dwReturned = 0;

	// Get the default printer (if any)
	wxString defPrinter;
	{
        TCHAR defaultBuff[256];
	    DWORD dBuffSize = sizeof(defaultBuff) / sizeof(TCHAR);

		if (::GetDefaultPrinter(defaultBuff, &dBuffSize))
		{
			defPrinter = defaultBuff;
		}
	}

	// First find out how much space we need for the return buffers
	if(!EnumPrinters(PRINTER_ENUM_CONNECTIONS | PRINTER_ENUM_LOCAL, 
		NULL, 4, 0, 0, &dwNeeded, &dwReturned)
        && GetLastError() != ERROR_INSUFFICIENT_BUFFER)
    {
        wxLogTrace("printing",
            "ERROR: cannot EnumPrinters [%ld]",GetLastError());

		return availablePrinters;
    }

	// Allocate the buffer
	PRINTER_INFO_4 *printerInfo = (PRINTER_INFO_4 *)malloc(dwNeeded);

	if (!printerInfo)
	{
		wxLogTrace("printing",
            "ERROR: can't get %ld bytes needed for PRINTER_INFO_1",
			dwNeeded);

		return availablePrinters;
	}

	if(!EnumPrinters(PRINTER_ENUM_CONNECTIONS | PRINTER_ENUM_LOCAL, NULL,
		4, (BYTE*)printerInfo, dwNeeded, &dwNeeded, &dwReturned))
	{
		wxLogTrace("printing",
            "ERROR: cannot EnumPrinters [%ld]", GetLastError());

		return availablePrinters;
	}

	for (int i = 0; i < (int)dwReturned; i++)
	{
	    PRINTER_INFO_4 *currPrinter = &printerInfo[i];

		wxString currPrinterName(currPrinter->pPrinterName);
		bool isDefault = (currPrinterName == defPrinter);

		wxSharedPtr<wxNativePrinterInfo> nativePI(
			new wxWindowsNativePrinterInfo(currPrinterName, isDefault));
		wxSharedPtr<wxPrinterInfo> pi(new wxPrinterInfo(nativePI));
		availablePrinters.push_back(pi);
	}

	free(printerInfo);

	return availablePrinters;
}

wxWindowsNativePrinterInfo::wxWindowsNativePrinterInfo(
       const wxString &printerName, bool isDefault) :
                     wxNativePrinterInfo(isDefault),
                     printerName(printerName)
{
	// Create the paper database
	TCHAR *printerNameTChar = const_cast<TCHAR*>(printerName.t_str());

	DWORD numPaperSizes = DeviceCapabilities(
		printerNameTChar, _T(""), DC_PAPERS, NULL, NULL);
	if(numPaperSizes < 0)
	{
		wxLogMessage(wxT("Could not get paper sizes for printer")); 
	}

	WORD *paperIDs = (WORD*) malloc(numPaperSizes * sizeof(WORD));
	LPWSTR paperNames = (LPWSTR) malloc(numPaperSizes * sizeof(wchar_t) * PAPERNAME_LENGTH);
	POINT *paperDims = (POINT*) malloc(numPaperSizes * sizeof(POINT));
	if(!paperIDs || !paperNames || !paperDims)
	{
		if(paperIDs)
		{
			free(paperIDs);
		}
		if(paperNames)
		{
			free(paperNames);
		}
		if(paperDims)
		{
			free(paperDims);
		}
		wxLogMessage(wxT("Out of memory")); 
		return;
	}


	DWORD ret = DeviceCapabilities(
		printerNameTChar, _T(""), DC_PAPERS, (LPWSTR)paperIDs, NULL);
	if(ret < 0)
	{
		wxLogMessage(wxT("Could not get paper IDs for printer")); 
		free(paperIDs);
		free(paperNames);
		free(paperDims);
		return;
	}
	ret = DeviceCapabilities(
		printerNameTChar, _T(""), DC_PAPERNAMES, paperNames, NULL);
	if(ret < 0)
	{
		wxLogMessage(wxT("Could not get paper names for printer")); 
		free(paperIDs);
		free(paperNames);
		free(paperDims);
		return;
	}
	ret = DeviceCapabilities(
		printerNameTChar, _T(""), DC_PAPERSIZE, (LPWSTR)paperDims, NULL);
	if(ret < 0)
	{
		wxLogMessage(wxT("Could not get paper dimensions for printer")); 
		free(paperIDs);
		free(paperNames);
		free(paperDims);
		return;
	}

	paperDatabase.reset(new wxPrintPaperDatabase());
	if(!paperDatabase)
	{
		wxLogMessage(wxT("Out of memory")); 
		free(paperIDs);
		free(paperNames);
		free(paperDims);
		return;
	}

	// Use the standard wxWidgets IDs where possible
	wxSharedPtr<wxPrintPaperDatabase> defaultPaperDatabase;
	defaultPaperDatabase.reset(new wxPrintPaperDatabase());
	if(!defaultPaperDatabase)
	{
		wxLogMessage(wxT("Out of memory")); 
		free(paperIDs);
		free(paperNames);
		free(paperDims);
		return;
	}
	defaultPaperDatabase->CreateDatabase();

	for(DWORD i = 0; i < numPaperSizes; ++i)
	{
		WORD winPaperID = paperIDs[i];
		wxPrintPaperType* paperType = defaultPaperDatabase->FindPaperTypeByPlatformId(winPaperID);
		wxPaperSize paperID;
		if(paperType)
		{
			paperID = paperType->GetId();
		}
		else
		{
			// Have a printer-specific paper type
			paperID = (wxPaperSize)(wxPAPER_USER + winPaperID);
			while(paperDatabase->FindPaperTypeByPlatformId(paperID))
			{
				// Find an unused ID
				paperID = (wxPaperSize)(paperID + 1);
			}
		}
		int width = paperDims[i].x;
		int height = paperDims[i].y;
		paperDatabase->AddPaperType(paperID, winPaperID, 
			wxString(&(paperNames[PAPERNAME_LENGTH * i]), PAPERNAME_LENGTH), 
			width, height);
	}

	free(paperIDs);
	free(paperNames);
	free(paperDims);
}

#define PAPERNAME_LENGTH 64

wxSharedPtr<wxPrintPaperDatabase> wxWindowsNativePrinterInfo::GetSupportedPaperSizes() const
{
	return paperDatabase;
}

void wxWindowsNativePrinterInfo::getNonPrintableMargins(
		unsigned int *leftMargin, unsigned int *rightMargin,
		unsigned int *topMargin, unsigned int *bottomMargin) const 
{
	TCHAR *printerNameTChar = const_cast<TCHAR*>(printerName.t_str());

    HDC hDC = CreateDC(NULL, printerNameTChar, NULL, NULL);

    if (!hDC)
    {
        wxLogLastError(wxT("CreateDC(printer)"));
    }

	unsigned int physOffsetX = 0;
	unsigned int physOffsetY = 0;
	physOffsetX = GetDeviceCaps(hDC, PHYSICALOFFSETX);
	physOffsetY = GetDeviceCaps(hDC, PHYSICALOFFSETY);
	unsigned int dpiX = GetDeviceCaps (hDC, LOGPIXELSX);
	unsigned int dpiY = GetDeviceCaps (hDC, LOGPIXELSY);
	unsigned int physWidth = GetDeviceCaps(hDC, PHYSICALWIDTH);
	unsigned int physHeight = GetDeviceCaps(hDC, PHYSICALHEIGHT);
	unsigned int printableWidth = GetDeviceCaps(hDC, HORZRES);
	unsigned int printableHeight = GetDeviceCaps(hDC, VERTRES);

	if(leftMargin)
	{
		*leftMargin = (unsigned int)INCHES_TO_TENTHSMM(physOffsetX / (double)dpiX);
	}

	if(topMargin)
	{
		*topMargin = (unsigned int)INCHES_TO_TENTHSMM(physOffsetY / (double)dpiY);
	}

	if(rightMargin)
	{
		*rightMargin = (unsigned int)INCHES_TO_TENTHSMM(
			(physWidth - printableWidth - physOffsetX) / (double)dpiX);
	}

	if(bottomMargin)
	{
		*bottomMargin = (unsigned int)INCHES_TO_TENTHSMM(
			(physHeight - printableHeight - physOffsetY) / (double)dpiY);
	}
    
	DeleteDC(hDC);
}

bool wxWindowsNativePrinterInfo::IsSamePrinter(const wxPrintData &printData) const
{
	return (printData.GetPrinterName() == printerName);
}

const wxPrintData wxWindowsNativePrinterInfo::GetPrintData()
{
	wxPrintData printData;

	printData.SetPrinterName(printerName);
	printData.ConvertToNative();

	return printData;
}

void wxWindowsNativePrinterInfo::UpdatePrintData(wxPrintData &printData) const
{
	printData.SetPrinterName(printerName);
	printData.ConvertToNative();
	printData.ConvertFromNative();

	if(printData.GetPaperId() == wxPAPER_NONE)
	{
		// Find the paper ID in our own database
		wxPrintPaperType* paperType = 
			paperDatabase->FindPaperType(printData.GetPaperSize() * 10);
		if(!paperType)
		{
			// Maybe DMPAPER_USER?
			paperType = paperDatabase->FindPaperTypeByPlatformId(DMPAPER_USER);
		}
		if(paperType)
		{
			// Setting the paper ID to that in our database for display
			printData.SetPaperId(paperType->GetId());
		}
	}
}

bool wxWindowsNativePrinterInfo::SetPrintDataPaperType(wxPaperSize paperID, wxPrintData &printData) const
{
	wxPrintPaperType* paperType = paperDatabase->FindPaperType(paperID);
	if(!paperType)
	{
		return false;
	}
	
	// Need to get wxWindowsPrintNativeData's m_customWindowsPaperId set appropriately for printer-specific
	// paper sizes that wxWidgets doesn't recognize
	printData.ConvertToNative();
	wxWindowsPrintNativeData* nativePrintData = (wxWindowsPrintNativeData*)printData.GetNativeData();
	HGLOBAL hDevMode = static_cast<HGLOBAL>(nativePrintData->GetDevMode());
	if(!hDevMode)
	{
		wxLogMessage(wxT("Could not get print devmode")); 
		return false;
	}
	GlobalPtrLock lockDevMode(hDevMode);
    DEVMODE* pDevMode = static_cast<DEVMODE *>(lockDevMode.Get());
	pDevMode->dmPaperSize = paperType->GetPlatformId();
	printData.ConvertFromNative();

	// Set both the ID and size
	printData.SetPaperId(paperID);
	wxSize paperDims = paperType->GetSizeMM();
	printData.SetPaperSize(paperDims);
	
	return true;
}

wxWindowsPrinterPrefsDialog::wxWindowsPrinterPrefsDialog(wxWindow *parent, wxPrintData *data)
{
	if(data)
	{
		m_printData = *data;
	}

	// Get the printer's paper database
	wxSharedPtr<wxPrinterInfo> printer = wxPrinterSysInfo::GetPrinterByPrintData(data);
	m_paperDatabase = printer->GetSupportedPaperSizes();

	m_dialogParent = parent;
}

wxWindowsPrinterPrefsDialog::~wxWindowsPrinterPrefsDialog()
{
	// Nothing to do here
}

int wxWindowsPrinterPrefsDialog::ShowModal()
{
	HWND parentHWnd;
	if (m_dialogParent)
	{
        parentHWnd = (HWND) m_dialogParent->GetHWND();
	}
    else if (wxTheApp->GetTopWindow())
	{
        parentHWnd = (HWND) wxTheApp->GetTopWindow()->GetHWND();
	}
    else
	{
        parentHWnd = 0;
	}

    HANDLE hPrinter;
    LPDEVMODE   pDevMode;
    DWORD       dwRet;
	TCHAR *printerName = const_cast<TCHAR*>(m_printData.GetPrinterName().t_str());

	m_printData.ConvertToNative();
	wxWindowsPrintNativeData* nativePrintData = (wxWindowsPrintNativeData*)m_printData.GetNativeData();
	HGLOBAL hDevMode = static_cast<HGLOBAL>(nativePrintData->GetDevMode());
	if(!hDevMode)
	{
		wxLogMessage(wxT("Could not get print devmode")); 
		return 0;
	}
	GlobalPtrLock lockDevMode(hDevMode);
    pDevMode = static_cast<DEVMODE *>(lockDevMode.Get());

	// WxWidgets doesn't support per-printer page sizes (it uses a single global page database),
	// so we use our database instead to fill in the dmPaperSize field
	if(pDevMode->dmFields & DM_PAPERSIZE)
	{
		wxPrintPaperType* paperType = m_paperDatabase->FindPaperType(m_printData.GetPaperId());
		if(!paperType)
		{
			// Try to get the paper type by size
			paperType = m_paperDatabase->FindPaperType(m_printData.GetPaperSize() * 10);
		}
		if(!paperType)
		{
			// Maybe DMPAPER_USER?
			paperType = m_paperDatabase->FindPaperTypeByPlatformId(DMPAPER_USER);
		}
		if(paperType)
		{
			pDevMode->dmPaperSize = paperType->GetPlatformId();
		}
		else
		{
			pDevMode->dmPaperSize = 0;
		}
	}

	if(m_printData.GetPaperId() == wxPAPER_NONE)
	{
		// WxWidgets uses mm for paper size with integers, so we lose precision. Ugh!
		wxSize paperSize = m_printData.GetPaperSize();
		pDevMode->dmPaperWidth = paperSize.GetWidth() * 10;
		pDevMode->dmPaperLength = paperSize.GetHeight() * 10;
		if(pDevMode->dmPaperWidth > 0 && pDevMode->dmPaperLength > 0)
		{
			pDevMode->dmFields |= DM_PAPERLENGTH | DM_PAPERWIDTH;
		}
	}


	// Open the printer
	if(!OpenPrinter(printerName, &hPrinter, 0))
	{
		wxLogMessage(wxT("Could not open the printer")); 
		return 0;
	}


	// Show the dialog
	int retVal;
	dwRet = DocumentProperties(parentHWnd,
        hPrinter,
        printerName,
        pDevMode,       /* Reuse our buffer for output. */ 
        pDevMode,       /* Pass the driver our changes. */ 
        DM_IN_BUFFER |  /* Commands to Merge our changes and */ 
		DM_IN_PROMPT |  /* Display the dialog and */
        DM_OUT_BUFFER); /* write the result. */ 
	if(dwRet == IDOK)
	{
			m_printData.ConvertFromNative();

			// Exctracting the paper details from the database, just in case we have a printer-specific 
			// size that wxWidgets doesn't support
			if(pDevMode->dmFields & DM_PAPERSIZE)
			{
				wxPrintPaperType* paperType = m_paperDatabase->FindPaperTypeByPlatformId(pDevMode->dmPaperSize);
				if(paperType)
				{
					m_printData.SetPaperId(paperType->GetId());
					m_printData.SetPaperSize(paperType->GetSizeMM());
				}
				else
				{
					m_printData.SetPaperId(wxPAPER_NONE);

					// WxWidgets uses mm for paper size with integers, so we lose precision. Ugh!
					wxSize paperSize(pDevMode->dmPaperWidth / 10, 
						pDevMode->dmPaperLength / 10); 
					m_printData.SetPaperSize(paperSize);
				}
			}

		retVal = wxID_OK;
	}
	else
	{
		retVal = wxID_CANCEL;
	}
	
	ClosePrinter(hPrinter);
	return retVal;
}

wxPrintData& wxWindowsPrinterPrefsDialog::GetPrintData()
{
	return m_printData;
}

#endif
